#include <Xmd/Buffer.h>
#include <stdlib.h>
#include <string.h>

#define XmdBUFFER_GROWTH_FACTOR 2

struct _XmdBuffer {
	void * data;
	size_t element;
	size_t length;
	size_t capacity;
};

static void  XmdBufferResize (XmdBuffer *buffer, size_t length);
static void  XmdBufferFit    (XmdBuffer *buffer);
static void *XmdBufferOffset (XmdBuffer *buffer, size_t index);

XmdBuffer *XmdBufferNew (size_t element) {
	XmdBuffer *buffer = calloc(1, sizeof(XmdBuffer));
	buffer->element  = element;
	buffer->capacity = 1;
	XmdBufferFit(buffer);
	return buffer;
}

void XmdBufferResize (XmdBuffer *buffer, size_t length) {
	buffer->length = length;
	XmdBufferFit(buffer);
}

void XmdBufferFit (XmdBuffer *buffer) {
	if (buffer->length > buffer->capacity) {
		buffer->capacity *= XmdBUFFER_GROWTH_FACTOR;
		buffer->data = realloc (
			buffer->data,
			buffer->capacity * buffer->element);
	}
}

void *XmdBufferOffset (XmdBuffer *buffer, size_t index) {
	return buffer->data + (index * buffer->element);
}

void XmdBufferPush (XmdBuffer *buffer, void *element) {
	XmdBufferResize(buffer, buffer->length + 1);
	memcpy (
		XmdBufferOffset(buffer, buffer->length - 1),
		element,
		buffer->element);
}

void *XmdBufferBreak (XmdBuffer *buffer) {
	void *data = buffer->data;
	free(buffer);
	return data;
}

void XmdBufferFree (XmdBuffer *buffer) {
	free(buffer->data);
	free(buffer);
}
